from .nextflow_building_blocks import Nextflow_Building_Blocks
from .bioflowinsighterror import BioFlowInsightError
from .root import Root
import re
from .outils import *

from . import constant



class Main(Nextflow_Building_Blocks):
    def __init__(self, code, nextflow_file):
        Nextflow_Building_Blocks.__init__(self, code, initialise_code=False)
        self.nextflow_file = nextflow_file
        self.initialised = False
        self.root = None 

    def get_order_execution_executors(self, dico, seen):
        executors = self.get_all_executors_in_subworkflow()
        pos = {}
        for e in executors:
            if(not e.get_artificial_status()):
                code = e.get_code(get_OG = True)
                #We don't have to check the calls -> since there are renamed with their ids when we rewrite the code -> so it solve the issue
                if(code in seen and e.get_type()=="Operation"):
                    raise BioFlowInsightError(f'Operation "{code}" appears twice in the workflow in the exact same way. BioFlow-Insight cannot rewrite the workflow then, try slighly changing how one of the executors is defined', type = "Rewrite Error")
                seen[code] = e
                pos[e] = e.get_position_in_main(e)
        pos =  {k: v for k, v in sorted(pos.items(), key=lambda item: item[1])}
        for e in pos:
            if(e.get_type()=="Call"):
                if(e.get_first_element_called().get_type()=="Subworkflow"):
                    dico[e] = {}
                    e.get_first_element_called().get_order_execution_executors(dico[e], seen)
                else:
                    dico[e.get_first_element_called()] = pos[e]
            else:
                dico[e] = pos[e]
        return dico

    #We do thus to defitiate the subworkflwow and main case
    def get_code_to_simplify(self):
        return self.get_code()

    
    
    def simplify_code(self):
        code = self.get_code_to_simplify()
        all_executors = self.get_all_executors_in_subworkflow()
        #We do this so that the longest operation and calls are rewritten first in the code -> to avoid problems
        executor_2_length = {}
        for e in all_executors:
            executor_2_length[e] = len(e.get_code(get_OG = True))
        sorted_executor_2_length = {k: v for k, v in sorted(executor_2_length.items(), key=lambda item: item[1], reverse=True)}
        
        for exe in sorted_executor_2_length:
            if(exe.get_type()=="Call" or exe.get_type()=="Operation"):
                old = exe.get_code(get_OG = True, remove_emit_and_take = True, replace_calls = False)
                new = exe.simplify_code(return_tab = False)
                if(new!=old):
                    temp = code
                    code = code.replace(old, new, 1)
                    if(temp==code):
                        print(exe)
                        print(code)
                        print("- old", f'"{old}"')
                        print("- new", f'"{new}"')
                        raise Exception("This shouldn't happen the Executor was not replaced")
            else:
                raise Exception("This shouldn't happen")
        return code

    def get_position_in_main(self, executor):
        code = self.get_code()
        return code.find(executor.get_code(get_OG = True))

    def get_string_line(self, bit_of_code):
        return self.nextflow_file.get_string_line(bit_of_code)
    
    #def check_in_channels(self, channel):
    #    return self.root.check_in_channels(channel)

    def get_conditions_2_ignore(self):
        return self.nextflow_file.get_conditions_2_ignore()

    def get_modules_defined(self):
        return self.nextflow_file.get_modules_defined()
    
    def get_duplicate_status(self):
        return self.nextflow_file.get_duplicate_status()
    
    def get_DSL(self):
        return self.nextflow_file.get_DSL()
    
    def get_file_address(self):
        return self.nextflow_file.get_file_address()


    def get_output_dir(self):
        return self.nextflow_file.get_output_dir()


    def get_type(self):
        return "Main"

    def get_all_calls_in_subworkflow(self):
        dico = {}
        self.root.get_all_calls_in_subworkflow(calls = dico)
        return list(dico.keys())
    
    def get_all_executors_in_subworkflow(self):
        dico = {}
        self.root.get_all_executors_in_subworkflow(calls = dico)
        return list(dico.keys())


    #TODO -> write tests to test this method
    def get_all_calls_in_workflow(self):
        all_calls = self.get_all_calls_in_subworkflow()
        dico = {}
        for c in all_calls:
            sub = c.get_first_element_called()
            if(sub.get_type()=="Subworkflow"):
                if(c not in dico):
                    sub_calls = sub.get_all_calls_in_workflow()
                    for sub_c in sub_calls:
                        dico[sub_c] = ""
        for c in all_calls:
            dico[c] = ""

        return list(dico.keys())
    
    #TODO -> write tests to test this method 
    def get_all_executors_in_workflow(self):
        all_executors = self.get_all_executors_in_subworkflow()
        dico = {}
        for e in all_executors:
            dico[e] = ""

        calls = self.get_all_calls_in_workflow()
        for call in calls:
            
            sub = call.get_first_element_called()
            if(sub.get_type()=="Subworkflow"):
                sub_calls = sub.get_all_executors_in_workflow()
                for sub_c in sub_calls:
                    dico[sub_c] = ""


        #for e in all_executors:
        #    if(e.get_type()=="Call"):
        #        for c in e.get_all_calls():
        #            sub = c.get_first_element_called()
        #            if(sub.get_type()=="Subworkflow"):
        #                if(c not in dico):
        #                    sub_calls = sub.get_all_executors_in_workflow()
        #                    for sub_c in sub_calls:
        #                        dico[sub_c] = ""
                    
        return list(dico.keys())

    
    def check_includes(self):
        code = self.get_code()

        pattern = constant.FULL_INCLUDE
        for match in re.finditer(pattern, code):
            if(self.get_type()=="Main"):
                raise BioFlowInsightError(f"An include ('{match.group(0)}') was found in the main in the file '{self.get_file_address()}'. FlowInsight does not support this -> see specification list.", type = 12,origin=self)
            elif(self.get_type()=="Subworkflow"):
                raise BioFlowInsightError(f"An include ('{match.group(0)}') was found in the subworkflow '{self.get_name()}' in the file '{self.get_file_address()}'. FlowInsight does not support this -> see specification list.", type = 12, origin=self)
            else:
                raise Exception("This shouldn't happen!")
            
        
    def initialise(self):
        if(not self.initialised):

            self.initialised=True

            #Get the modules (Processes defined for the main/subworkflow)
            self.modules_defined = self.nextflow_file.get_modules_defined()

            #Check that includes are not defined in the main or subworkflows
            self.check_includes()
            self.root = Root(code=self.get_code(), origin=self, modules_defined=self.modules_defined, subworkflow_inputs = [])
            #The weird DSL1 bug is here
            #self.root = Root(code=self.get_code(), origin=self, modules_defined=self.modules_defined)
            self.root.initialise()




    def get_structure(self, dico):
        self.root.get_structure(dico)
        return dico
    
    def get_most_influential_conditions(self):
        most_influential_conditions = {}
        if(self.nextflow_file.get_DSL()=="DSL2"):
            all_calls = self.get_all_calls_in_subworkflow()
            for c in all_calls:
                if(c.get_first_element_called().get_type()=="Process"):
                    for condition in c.get_block().get_all_conditions(conditions = {}):
                        try:
                            temp = most_influential_conditions[condition]
                        except:
                            most_influential_conditions[condition] = 0
                        most_influential_conditions[condition]+=1
                if(c.get_first_element_called().get_type()=="Subworkflow"):
                    #Adding the number of calls from a process at the root of the subworkflow
                    num = 0
                    calls_at_root = c.get_first_element_called().root.get_calls_same_level()
                    for call_at_root in calls_at_root:
                        for c_at_root in call_at_root.get_all_calls():
                            if(c_at_root.get_first_element_called().get_type()=="Process"):
                                num +=1

                    most_influential_conditions_in_sub = c.get_first_element_called().get_most_influential_conditions()
                    #Adding the conditions from inside the subworkflow
                    for condition in most_influential_conditions_in_sub:
                        try:
                            temp = most_influential_conditions[condition]
                        except:
                            most_influential_conditions[condition] = 0
                        num+=most_influential_conditions_in_sub[condition]
                        most_influential_conditions[condition]+=most_influential_conditions_in_sub[condition]
                    #Adding calls from the subworkflow to the conditions
                    for condition in c.get_block().get_all_conditions(conditions = {}):
                        try:
                            temp = most_influential_conditions[condition]
                        except:
                            most_influential_conditions[condition] = 0
                        most_influential_conditions[condition]+=num
            return most_influential_conditions
        elif(self.nextflow_file.get_DSL()=="DSL1"):
            modules = self.get_modules_defined()
            for p in modules:
                if(p.get_type()=="Process"):
                    if(p.is_initialised()):
                        for condition in p.origin.get_all_conditions(conditions = {}):
                            try:
                                temp = most_influential_conditions[condition]
                            except:
                                most_influential_conditions[condition] = 0
                            most_influential_conditions[condition]+=1
            return most_influential_conditions
        else:
            raise Exception("This shouldn't happen")